import dataclasses

import numpy as np
import numpy.testing as npt
import pytest

from caimira.calculator.models import models


@pytest.fixture
def baseline_slidingwindow(data_registry):
    return models.SlidingWindow(
            data_registry=data_registry,
            active=models.SpecificInterval(((0, 4), (5, 9))),
            outside_temp=models.PiecewiseConstant((0, 24), (283,)),
            window_height=1.6, opening_length=0.6,
        )


@pytest.fixture
def baseline_hingedwindow():
    return models.HingedWindow(
            active=models.SpecificInterval(((0, 4), (5, 9))),
            outside_temp=models.PiecewiseConstant((0, 24), (283,)),
            window_height=1.6, opening_length=0.6, window_width=1.,
        )


def test_number_of_windows(baseline_slidingwindow):
    room = models.Room(volume=75, inside_temp=models.PiecewiseConstant((0, 24), (293,)))
    two_windows = dataclasses.replace(baseline_slidingwindow, number_of_windows=2)

    one_window_exchange = baseline_slidingwindow.air_exchange(room, 1)
    two_window_exchange = two_windows.air_exchange(room, 1)
    assert one_window_exchange != 0
    assert one_window_exchange * 2 == two_window_exchange


@pytest.mark.parametrize(
    "window_width, expected_discharge_coefficient",
    [
        [0.5, 0.447],
        [1., 0.379],
        [2., 0.328],
        [4., 0.308],
    ],
)
def test_hinged_window(baseline_hingedwindow, window_width,
                       expected_discharge_coefficient):
    hinged_window = dataclasses.replace(baseline_hingedwindow,
                                        window_width=window_width)

    npt.assert_allclose(hinged_window.discharge_coefficient,
                        expected_discharge_coefficient, rtol=1e-2)


@pytest.mark.parametrize(
    "override_params", [
        {'window_height': np.array([0.15, 0.20, 0.25])},
        {'window_width': np.array([0.15, 0.20, 0.25])},
        {'opening_length': np.array([0.15, 0.20, 0.25])},
        {'outside_temp': models.PiecewiseConstant(
            (0, 2, 3), (np.array([20, 30, 28]), np.array([25, 30, 27]))
        )},
    ]
)
def test_hinged_window_vectorisation(override_params):
    defaults = {
        'window_height': 0.15,
        'window_width': 0.15,
        'opening_length': 0.15,
        'outside_temp': models.PiecewiseConstant((0, 2, 3), (10, 15)),
    }
    defaults.update(override_params)
    room = models.Room(volume=75, inside_temp=models.PiecewiseConstant((0, 2, 3), (20, 25)))
    t = 0.5
    window = models.HingedWindow(models.PeriodicInterval(60, 30), **defaults)
    if {'window_height', 'opening_length', 'window_width'}.intersection(override_params):
        assert isinstance(window.discharge_coefficient, np.ndarray)
    else:
        assert isinstance(window.discharge_coefficient, float)

    assert isinstance(window.air_exchange(room, t), np.ndarray)


def test_sliding_window(baseline_slidingwindow):
    assert baseline_slidingwindow.discharge_coefficient == 0.6


def test_hvac_mechanical_vectorisation():
    room = models.Room(volume=50)
    interval = models.SpecificInterval(((0, 4), (5, 9)))
    t = 0.5
    q_air_mech = np.array([250., 500.])
    v = models.HVACMechanical(interval,q_air_mech)
    assert isinstance(v.air_exchange(room, t), np.ndarray)
    npt.assert_array_equal(v.air_exchange(room, t),
                        np.array([250/room.volume, 500/room.volume]))


def test_hepa_filter_vectorisation():
    room = models.Room(volume=50)
    interval = models.SpecificInterval(((0, 4), (5, 9)))
    t = 0.5
    q_air_mech = np.array([250., 500.])
    v = models.HEPAFilter(interval,q_air_mech)
    assert isinstance(v.air_exchange(room, t), np.ndarray)
    npt.assert_array_equal(v.air_exchange(room, t),
                        np.array([250/room.volume, 500/room.volume]))


def test_multiple(baseline_slidingwindow, baseline_hingedwindow):
    v = models.MultipleVentilation([baseline_hingedwindow, baseline_slidingwindow])
    room = models.Room(75)
    t = 1
    assert v.air_exchange(room, t) == (
            baseline_slidingwindow.air_exchange(room, t) +
            baseline_hingedwindow.air_exchange(room, t)
    )


def test_multiple_vectorisation():
    interval = models.SpecificInterval(((0, 4), (5, 9)))
    v1 = models.AirChange(interval, np.arange(10))
    v2 = models.AirChange(interval, np.arange(5))
    v3 = models.AirChange(interval, 10)

    room = models.Room(75)
    t_active = 2
    t_inactive = 4.5

    assert models.MultipleVentilation([v1, v2]).air_exchange(room, t_inactive) == 0
    with pytest.raises(ValueError, match='operands could not be broadcast together'):
        models.MultipleVentilation([v1, v2]).air_exchange(room, t_active)

    r = models.MultipleVentilation([v2, v3]).air_exchange(room, t_active)
    assert isinstance(r, np.ndarray)
    np.testing.assert_array_equal(r, [10, 11, 12, 13, 14])



import typing

from caimira.calculator.validators.co2.co2_validator import CO2FormData
from caimira.calculator.store.data_registry import DataRegistry
from caimira.calculator.models.models import CO2DataModel


def generate_form_obj(form_data: typing.Dict, data_registry: DataRegistry) -> CO2FormData:
    return CO2FormData.from_dict(form_data=form_data, data_registry=data_registry)


def generate_model(form_obj: CO2FormData) -> CO2DataModel:
    return form_obj.build_model()


def generate_report(model: CO2DataModel) -> typing.Dict:
    return dict(model.CO2_fit_params())


def submit_CO2_form(form_data: typing.Dict) -> typing.Dict:
    data_registry: DataRegistry = DataRegistry()

    form_obj: CO2FormData = generate_form_obj(form_data=form_data, data_registry=data_registry)
    model: CO2DataModel = generate_model(form_obj=form_obj)
    report_data: typing.Dict = generate_report(model=model)

    return report_data



import numpy as np
import numpy.testing as npt
import pytest
from retry import retry

import caimira.calculator.models.monte_carlo as mc
from caimira.calculator.models import models, data
from caimira.calculator.models.monte_carlo.data import activity_distributions, virus_distributions, expiration_distributions, infectious_dose_distribution, viable_to_RNA_ratio_distribution
from caimira.calculator.validators.virus.virus_validator import build_expiration

SAMPLE_SIZE = 500_000
TOLERANCE = 0.05

# Load the weather data (temperature in kelvin) for Toronto.
toronto_coordinates = (43.667, 79.400)
toronto_hourly_temperatures_celsius_per_hour = data.get_hourly_temperatures_celsius_per_hour(
    toronto_coordinates)

# Toronto hourly temperatures as piecewise constant function (in Kelvin).
TorontoTemperatures_hourly = {
    month: models.PiecewiseConstant(
        # NOTE:  It is important that the time type is float, not np.float, in
        # order to allow hashability (for caching).
        tuple(float(time) for time in range(25)),
        tuple(273.15 + np.array(temperatures)),
    )
    for month, temperatures in toronto_hourly_temperatures_celsius_per_hour.items()
}

# Same Toronto temperatures on a finer temperature mesh (every 6 minutes).
TorontoTemperatures = {
    month: TorontoTemperatures_hourly[month].refine(refine_factor=10)
    for month, temperatures in toronto_hourly_temperatures_celsius_per_hour.items()
}


# References values for infection_probability and expected new cases
# in the following tests, were obtained from the feature/mc branch
@pytest.fixture
def shared_office_mc(data_registry):
    """
    Corresponds to the 1st line of Table 4 in https://doi.org/10.1101/2021.10.14.21264988
    """
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=50, inside_temp=models.PiecewiseConstant((0., 24.), (298,)), humidity=0.5),
        ventilation=models.MultipleVentilation(
            ventilations=(
                models.SlidingWindow(
                    data_registry=data_registry,
                    active=models.PeriodicInterval(period=120, duration=120),
                    outside_temp=data.GenevaTemperatures['Jun'],
                    window_height=1.6,
                    opening_length=0.2,
                ),
                models.AirChange(active=models.PeriodicInterval(period=120, duration=120), air_exch=0.25),
            )
        ),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=mc.SpecificInterval(present_times=((0, 3.5), (4.5, 9))),
            virus=virus_distributions(data_registry)['SARS_CoV_2_DELTA'],
            mask=models.Mask.types['No mask'],
            activity=activity_distributions(data_registry)['Seated'],
            expiration=build_expiration(data_registry, {'Speaking': 0.33, 'Breathing': 0.67}),
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=3,
            presence=mc.SpecificInterval(present_times=((0, 3.5), (4.5, 9))),
            activity=activity_distributions(data_registry)['Seated'],
            mask=models.Mask.types['No mask'],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def classroom_mc(data_registry):
    """
    Corresponds to the 2nd line of Table 4 in https://doi.org/10.1101/2021.10.14.21264988
    """
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=160, inside_temp=models.PiecewiseConstant((0., 24.), (293,)), humidity=0.3),
        ventilation=models.MultipleVentilation(
            ventilations=(
                models.SlidingWindow(
                    data_registry=data_registry,
                    active=models.PeriodicInterval(period=120, duration=120),
                    outside_temp=TorontoTemperatures['Dec'],
                    window_height=1.6,
                    opening_length=0.2,
                ),
                models.AirChange(active=models.PeriodicInterval(period=120, duration=120), air_exch=0.25),
            )
        ),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=models.SpecificInterval(((0, 2), (2.5, 4), (5, 7), (7.5, 9))),
            virus=virus_distributions(data_registry)['SARS_CoV_2_ALPHA'],
            mask=models.Mask.types["No mask"],
            activity=activity_distributions(data_registry)['Light activity'],
            expiration=build_expiration(data_registry, 'Speaking'),
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=19,
            presence=models.SpecificInterval(((0, 2), (2.5, 4), (5, 7), (7.5, 9))),
            activity=activity_distributions(data_registry)['Seated'],
            mask=models.Mask.types["No mask"],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def ski_cabin_mc(data_registry):
    """
    Corresponds to the 3rd line of Table 4 in https://doi.org/10.1101/2021.10.14.21264988
    """
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=10, humidity=0.3),
        ventilation=models.MultipleVentilation(
            (models.AirChange(active=models.PeriodicInterval(period=120, duration=120), air_exch=0.0),
            models.AirChange(active=models.PeriodicInterval(period=120, duration=120), air_exch=0.25))),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=models.SpecificInterval(((0, 20/60),)),
            virus=virus_distributions(data_registry)['SARS_CoV_2_DELTA'],
            mask=models.Mask.types['No mask'],
            activity=activity_distributions(data_registry)['Moderate activity'],
            expiration=build_expiration(data_registry, 'Speaking'),
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=3,
            presence=models.SpecificInterval(((0, 20/60),)),
            activity=activity_distributions(data_registry)['Moderate activity'],
            mask=models.Mask.types['No mask'],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def skagit_chorale_mc(data_registry):
    """
    Corresponds to the 4th line of Table 4 in https://doi.org/10.1101/2021.10.14.21264988,
    assuming viral is 10**9 instead of a LogCustomKernel distribution.
    """
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=810, humidity=0.5),
            ventilation=models.AirChange(
                active=models.PeriodicInterval(period=120, duration=120),
                air_exch=0.7),
            infected=mc.InfectedPopulation(
                data_registry=data_registry,
                number=1,
                presence=models.SpecificInterval(((0, 2.5), )),
                virus=mc.SARSCoV2(
                    viral_load_in_sputum=10**9,
                    infectious_dose=infectious_dose_distribution(),
                    viable_to_RNA_ratio=viable_to_RNA_ratio_distribution(),
                    transmissibility_factor=1.,
                ),
                mask=models.Mask.types['No mask'],
                activity=activity_distributions(data_registry)['Moderate activity'],
                expiration=build_expiration(data_registry, 'Shouting'),
                host_immunity=0.,
            ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=60,
            presence=models.SpecificInterval(((0, 2.5), )),
            activity=activity_distributions(data_registry)['Moderate activity'],
            mask=models.Mask.types['No mask'],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def bus_ride_mc(data_registry):
    """
    Corresponds to the 5th line of Table 4 in https://doi.org/10.1101/2021.10.14.21264988,
    assuming viral is 5*10**8 instead of a LogCustomKernel distribution.
    """
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=45, humidity=0.5),
            ventilation=models.AirChange(
                active=models.PeriodicInterval(period=120, duration=120),
                air_exch=1.25),
            infected=mc.InfectedPopulation(
                data_registry=data_registry,
                number=1,
                presence=models.SpecificInterval(((0, 1.67), )),
                virus=mc.SARSCoV2(
                    viral_load_in_sputum=5*10**8,
                    infectious_dose=infectious_dose_distribution(),
                    viable_to_RNA_ratio=viable_to_RNA_ratio_distribution(),
                    transmissibility_factor=1.,
                ),
                mask=models.Mask.types['No mask'],
                activity=activity_distributions(data_registry)['Seated'],
                expiration=build_expiration(data_registry, 'Speaking'),
                host_immunity=0.,
            ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=67,
            presence=models.SpecificInterval(((0, 1.67), )),
            activity=activity_distributions(data_registry)['Seated'],
            mask=models.Mask.types['No mask'],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def gym_mc(data_registry):
    """
    Gym model for testing
    """
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=300, humidity=0.5),
        ventilation=models.AirChange(
            active=models.SpecificInterval(((0., 24.),)),
            air_exch=6,
        ),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=2,
            virus=virus_distributions(data_registry)['SARS_CoV_2_ALPHA'],
            presence=mc.SpecificInterval(((0., 1.),)),
            mask=models.Mask.types["No mask"],
            activity=activity_distributions(data_registry)['Heavy exercise'],
            expiration=expiration_distributions(data_registry)['Breathing'],
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=28,
            presence=concentration_mc.infected.presence,
            activity=models.Activity.types['Heavy exercise'],
            mask=concentration_mc.infected.mask,
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def waiting_room_mc(data_registry):
    """
    Waiting room model for testing
    """
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=100, humidity=0.5),
        ventilation=models.AirChange(
            active=models.SpecificInterval(((0., 24.),)),
            air_exch=0.25,
        ),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            virus=virus_distributions(data_registry)['SARS_CoV_2_ALPHA'],
            presence=mc.SpecificInterval(((0., 2.),)),
            mask=models.Mask.types["No mask"],
            activity=activity_distributions(data_registry)['Seated'],
            expiration=build_expiration(data_registry, {'Speaking': 0.3, 'Breathing': 0.7}),
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=14,
            presence=concentration_mc.infected.presence,
            activity=models.Activity.types['Seated'],
            mask=concentration_mc.infected.mask,
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


# In the following tests all the initial expected values for emission
# rate were multiplied by the average of the distribution of fraction
# of infected viruses (0.305, for a uniform distribution from 0.01 to 0.6)
# following the change of convention that this ratio should be
# applied at the emission level.
@retry(tries=10)
@pytest.mark.parametrize(
    "mc_model, expected_pi, expected_new_cases, expected_dose, expected_ER_per_person",
    [
        ["shared_office_mc", 5.38, 0.16, 3.350, 1056*0.305],
        ["classroom_mc",     8.21, 1.56, 11.356, 7416*0.305],
        ["ski_cabin_mc",     12.92, 0.39, 21.796, 10231*0.305],
        ["skagit_chorale_mc",61.01, 36.53, 84.730, 190422*0.305],
        ["bus_ride_mc",      10.59, 7.06, 6.650, 5419*0.305],
        ["gym_mc",           0.52, 0.14, 0.249, 1450*0.305/2.], # there are two infected in this case
        ["waiting_room_mc",  1.53, 0.21, 0.844, 929*0.305],
    ]
)
def test_report_models(mc_model, expected_pi, expected_new_cases,
                       expected_dose, expected_ER_per_person, request):
    mc_model = request.getfixturevalue(mc_model)
    exposure_model = mc_model.build_model(size=SAMPLE_SIZE)
    npt.assert_allclose(exposure_model.infection_probability().mean(),
                        expected_pi, rtol=TOLERANCE)
    npt.assert_allclose(exposure_model.expected_new_cases().mean(),
                        expected_new_cases, rtol=TOLERANCE)
    npt.assert_allclose(exposure_model.deposited_exposure().mean(),
                        expected_dose, rtol=TOLERANCE)
    npt.assert_allclose(
        exposure_model.concentration_model.infected.emission_rate_per_person_when_present().mean(),
        expected_ER_per_person, rtol=TOLERANCE)


@retry(tries=10)
@pytest.mark.parametrize(
    "mask_type, month, expected_pi, expected_dose, expected_ER",
    [
        ["No mask", "Jul", 7.689, 10.050, 1034.435*0.305],
        ["Type I",  "Jul", 1.663, 0.938, 193.52*0.305],
        ["FFP2",    "Jul", 0.523, 0.253, 193.52*0.305],
        ["Type I",  "Feb", 0.659, 0.325, 193.52*0.305],
        ["Cloth",   "Feb", 2.653, 1.741, 673.10*0.305],
        ["Cloth",   "Jul", 5.322, 5.064, 673.10*0.305],
    ],
)
def test_small_shared_office_Geneva(data_registry, mask_type, month, expected_pi,
                                    expected_dose, expected_ER):
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=33, inside_temp=models.PiecewiseConstant((0., 24.), (293,)), humidity=0.5),
        ventilation=models.MultipleVentilation(
            (
                models.SlidingWindow(
                    data_registry=data_registry,
                    active=models.SpecificInterval(((0., 24.),)),
                    outside_temp=data.GenevaTemperatures[month],
                    window_height=1.5, opening_length=0.2,
                ),
                models.AirChange(
                    active=models.SpecificInterval(((0., 24.),)),
                    air_exch=0.25,
                ),
            ),
        ),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            virus=virus_distributions(data_registry)['SARS_CoV_2_ALPHA'],
            presence=mc.SpecificInterval(((9., 10+2/3), (10+5/6, 12.5), (13.5, 15+2/3), (15+5/6, 18.))),
            mask=models.Mask.types[mask_type],
            activity=activity_distributions(data_registry)['Seated'],
            expiration=build_expiration(data_registry, {'Speaking': 0.33, 'Breathing': 0.67}),
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    exposure_mc = mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=1,
            presence=concentration_mc.infected.presence,
            activity=activity_distributions(data_registry)['Seated'],
            mask=concentration_mc.infected.mask,
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )
    exposure_model = exposure_mc.build_model(size=SAMPLE_SIZE)
    npt.assert_allclose(exposure_model.infection_probability().mean(),
                        expected_pi, rtol=TOLERANCE)
    npt.assert_allclose(exposure_model.deposited_exposure().mean(),
                        expected_dose, rtol=TOLERANCE)
    npt.assert_allclose(
        exposure_model.concentration_model.infected.emission_rate_per_person_when_present().mean(),
        expected_ER, rtol=TOLERANCE)



import numpy as np
import pytest

import caimira.calculator.models.models


@pytest.mark.parametrize(
    "override_params", [
        {'viral_load_in_sputum': np.array([5e8, 1e9])},
        {'exhalation_rate': np.array([0.75, 0.81])},
    ]
)
def test_infected_population_vectorisation(override_params, data_registry):
    defaults = {
        'viral_load_in_sputum': 1e9,
        'exhalation_rate': 0.75,
    }
    defaults.update(override_params)

    office_hours = caimira.calculator.models.models.SpecificInterval(present_times=[(8,17)])
    infected = caimira.calculator.models.models.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=office_hours,
            mask=caimira.calculator.models.models.Mask(
                factor_exhale=0.95,
                η_inhale=0.3,
            ),
            activity=caimira.calculator.models.models.Activity(
                0.51,
                defaults['exhalation_rate'],
            ),
            virus=caimira.calculator.models.models.SARSCoV2(
                viral_load_in_sputum=defaults['viral_load_in_sputum'],
                infectious_dose=50.,
                viable_to_RNA_ratio = 0.5,
                transmissibility_factor=1.0,
            ),
            expiration=caimira.calculator.models.models._ExpirationBase.types['Breathing'],
            host_immunity=0.,
    )
    emission_rate = infected.emission_rate(10)
    assert isinstance(emission_rate, np.ndarray)
    assert emission_rate.shape == (2, )



from dataclasses import dataclass
import typing

import numpy as np
from scipy.integrate import quad
from scipy.special import erf
import numpy.testing as npt
import pytest
from retry import retry

import caimira.calculator.models.monte_carlo as mc
from caimira.calculator.models import models
from caimira.calculator.models.utils import method_cache
from caimira.calculator.models.models import _VectorisedFloat,Interval,SpecificInterval
from caimira.calculator.models.monte_carlo.data import (expiration_distributions,
        expiration_BLO_factors,short_range_expiration_distributions,
        short_range_distances,virus_distributions,activity_distributions)


SAMPLE_SIZE = 1_000_000
TOLERANCE = 0.04

sqrt2pi = np.sqrt(2.*np.pi)
sqrt2 = np.sqrt(2.)
ln2 = np.log(2)


@dataclass(frozen=True)
class SimpleConcentrationModel:
    """
    Simple model for the background (long-range) concentration, without
    all the flexibility of caimira.models.ConcentrationModel.
    For independent, end-to-end testing purposes.
    This assumes no mask wearing, and the same ventilation rate at all
    times.
    """

    #: Infected people presence interval
    infected_presence: Interval

    #: Viral load (RNA copies  / mL)
    viral_load: _VectorisedFloat

    #: Breathing rate (m^3/h)
    breathing_rate: _VectorisedFloat

    #: Room volume (m^3)
    room_volume: _VectorisedFloat

    #: Ventilation rate (air changes per hour) - including HEPA
    lambda_ventilation: _VectorisedFloat

    #: BLO factors
    BLO_factors: typing.Tuple[float, float, float]

    #: Number of infected people
    num_infected: int = 1

    #: Viable to RNA ratio
    viable_to_RNA: _VectorisedFloat = 0.5

    #: Host immunity factor (0. for not immune)
    HI: _VectorisedFloat = 0.

    #: Relative humidity RH
    humidity: float = 0.3

    #: Minimum particle diameter considered (microns)
    diameter_min: float = 0.1

    #: Maximum particle diameter considered (microns)
    diameter_max: float = 30.

    #: Evaporation factor
    evaporation: float = 0.3

    #: cn (cm^-3) for resp. the B, L and O modes. Corresponds to the
    # total concentration of aerosols for each mode.
    cn: typing.Tuple[float, float, float] = (0.06, 0.2, 0.0010008)

    # Mean of the underlying normal distributions (represents the log of a
    # diameter in microns), for resp. the B, L and O modes.
    mu: typing.Tuple[float, float, float] = (0.989541, 1.38629, 4.97673)

    # Std deviation of the underlying normal distribution, for resp.
    # the B, L and O modes.
    sigma: typing.Tuple[float, float, float] = (0.262364, 0.506818, 0.585005)

    def removal_rate(self) -> _VectorisedFloat:
        """
        Removal rate lambda in h^-1, excluding the deposition rate.
        """
        hl_calc = ((ln2/((0.16030 + 0.04018*(((293-273.15)-20.615)/10.585)
                                       +0.02176*(((self.humidity*100)-45.235)/28.665)
                                       -0.14369
                                       -0.02636*((293-273.15)-20.615)/10.585)))/60)

        return (self.lambda_ventilation
                + ln2/(np.where(hl_calc <= 0, 6.43, np.minimum(6.43, hl_calc))))

    def fraction_of_infectious_virus(self) -> _VectorisedFloat:
        """
        The fraction of infectious virus.
        """
        return self.viable_to_RNA * (1 - self.HI)

    @method_cache
    def deposition_removal_coefficient(self) -> float:
        """
        Coefficient in front of gravitational deposition rate, in h^-1.microns^-2
        Note: 0.4512 = 1.88e-4 * 3600 / 1.5
        """
        return 0.4512*(self.evaporation/2.5)**2

    @method_cache
    def aerosol_volume(self,diameter: float) -> float:
        """
        Particle volume in microns^3
        """
        return 4*np.pi/3. * (diameter/2.)**3

    @method_cache
    def Np(self,diameter: float,
           BLO_factors: typing.Tuple[float, float, float]) -> float:
        """
        Number of emitted particles per unit volume (BLO model)
        in cm^-3.ln(micron)^-1
        """
        result = 0.
        for cn,mu,sigma,famp in zip(self.cn,self.mu,self.sigma,
                                    BLO_factors):
            result += ( (cn * famp)/sigma *
                        np.exp(-(np.log(diameter)-mu)**2/(2*sigma**2)))
        return result/(diameter*sqrt2pi)

    def vR(self,diameter: float) -> float:
        """
        Emission rate per unit diameter, in RNA copies / h / micron
        """
        return (self.Np(diameter, self.BLO_factors)
                * self.aerosol_volume(diameter) * 1e-6)

    @method_cache
    def f(self, removal_rate: _VectorisedFloat, deltat: float) -> _VectorisedFloat:
        """
        A general function to compute the main integral over diameters
        """
        def integrand(diameter):
            # function to return the integrand
            a = self.deposition_removal_coefficient()
            a_dsquare = a*diameter**2
            return (self.vR(diameter)/(a_dsquare + removal_rate)
                    * np.exp(-a_dsquare*deltat))

        return (quad(integrand,self.diameter_min,self.diameter_max,
                     epsabs=0.,limit=500)[0]
                * self.viral_load * self.breathing_rate)

    def concentration(self,t: float) -> _VectorisedFloat:
        """
        Concentration at a given time t
        """
        trans_times = sorted(self.infected_presence.transition_times())
        if t==trans_times[0]:
            return 0.

        lambda_rate = self.removal_rate()
        # transition_times[i] < t <=  transition_times[i]+1
        i: int = np.searchsorted(trans_times,t) - 1 # type: ignore
        ti = trans_times[i]
        Pim1 = (False if i==0 else
            self.infected_presence.triggered((trans_times[i-1]+ti)/2.))
        Pi = self.infected_presence.triggered((ti+trans_times[i+1])/2.)

        result = (0 if not Pim1 else self.f(lambda_rate,t-ti))
        result -= (0 if not Pi else self.f(lambda_rate,t-ti))

        for k,tk in enumerate(trans_times[:i]):
            Pkm1 = (False if k==0 else
                self.infected_presence.triggered((trans_times[k-1]+tk)/2.))
            Pk = self.infected_presence.triggered((tk+trans_times[k+1])/2.)
            s = np.sum([lambda_rate*(trans_times[l]-trans_times[l-1])
                        for l in range(k+1,i+1)])
            result += ( (0 if not Pkm1 else self.f(lambda_rate,t-tk))
                       -(0 if not Pk else self.f(lambda_rate,t-tk))
                       ) * np.exp(-s)

        return ( ( (0 if not self.infected_presence.triggered(t)
                    else self.f(lambda_rate,0))
                  + result * np.exp(-lambda_rate*(t-ti)) )
                * self.num_infected * self.fraction_of_infectious_virus()
                / self.room_volume)


@dataclass(frozen=True)
class SimpleShortRangeModel:
    """
    Simple model for the short-range concentration, without
    all the flexibility of caimira.models.ShortRangeModel.
    For independent, end-to-end testing purposes.
    This assumes no mask wearing.
    """

    #: Time intervals in which a short-range interaction occurs
    interaction_interval: SpecificInterval

    #: Tuple with interpersonal distanced from infected person (m)
    distance : _VectorisedFloat = 0.854

    #: Breathing rate (m^3/h)
    breathing_rate: _VectorisedFloat = 0.51

    #: Exhalation coefficient
    φ = 2

    #: Tuple with BLO factors
    BLO_factors: typing.Tuple[float, float, float] = (1,0,0)

    #: Minimum diameter for integration (short-range only) (microns)
    diameter_min: float = 0.1

    #: Maximum diameter for integration (short-range only) (microns)
    diameter_max: float = 100.

    #: Average mouth opening diameter (m)
    mouth_diameter: float = 0.02

    #: Duration of the expiration period(s), assuming a 4s breath-cycle
    tstar: float = 2.

    #: Streamwise and radial penetration coefficients
    𝛽r1: float = 0.18
    𝛽r2: float = 0.2
    𝛽x1: float = 2.4

    @method_cache
    def dilution_factor(self) -> _VectorisedFloat:
        """
        Computes dilution factor at a certain distance x
        based on Wei JIA matlab script.
        """
        x = np.array(self.distance)
        dilution = np.empty(x.shape, dtype=np.float64)
        # Exhalation airflow, as per Jia et al. (2022), m^3/s
        Q_exh: _VectorisedFloat = self.φ * np.array(self.breathing_rate/3600)
        # The expired flow velocity at the noozle (mouth opening), m/s
        u0 = np.array(Q_exh/(np.pi/4. * self.mouth_diameter**2))
        # Parameters in the jet-like stage
        # position of virtual origin
        x0 = self.mouth_diameter/2/self.𝛽r1
        # Time of virtual origin
        t0 = (x0/self.𝛽x1)**2 * (Q_exh*u0)**(-0.5)
        # Transition point (in m)
        xstar = np.array(self.𝛽x1*(Q_exh*u0)**0.25*(self.tstar + t0)**0.5 - x0)
        # Dilution factor at the transition point xstar
        Sxstar = np.array(2.*self.𝛽r1*(xstar+x0)/self.mouth_diameter)

        # Calculate dilution factor at the short-range distance x
        dilution[x <= xstar] = 2.*self.𝛽r1*(x[x <= xstar] + x0)/self.mouth_diameter
        dilution[x > xstar] = Sxstar[x > xstar]*(1. + self.𝛽r2*(x[x > xstar]
                                - xstar[x > xstar])
                                /self.𝛽r1/(xstar[x > xstar] + x0))**3

        return dilution

    def jet_concentration(self,conc_model: SimpleConcentrationModel) -> _VectorisedFloat:
        """
        Virion concentration at the origin of the jet (close to
        the mouth of the infected person), in m^-3
        we perform the integral of Np(d)*V(d) over diameter analytically
        """
        vl = conc_model.viral_load
        viable_to_RNA = conc_model.viable_to_RNA
        dmin = self.diameter_min
        dmax = self.diameter_max
        result = 0.
        for cn,mu,sigma,famp in zip(conc_model.cn,conc_model.mu,conc_model.sigma,
                                    self.BLO_factors):
            d0 = np.exp(mu)
            ymin = (np.log(dmin)-mu)/(sqrt2*sigma)-3.*sigma/sqrt2
            ymax = (np.log(dmax)-mu)/(sqrt2*sigma)-3.*sigma/sqrt2
            result += ( (cn * famp * d0**3)/2. * np.exp(9*sigma**2/2.) *
                        (erf(ymax) - erf(ymin)) )
        return vl * viable_to_RNA * 1e-6 * result * np.pi/6.

    def concentration(self, conc_model: SimpleConcentrationModel, time: float) -> _VectorisedFloat:
        """
        Compute the short-range part of the concentration, and add it
        to the long-range concentration
        """
        if self.interaction_interval.triggered(time):
            lr_concentration = conc_model.concentration(time)
            S = self.dilution_factor()
            return (self.jet_concentration(conc_model)
                    - lr_concentration) / S
        else:
            return 0.


@dataclass(frozen=True)
class SimpleExposureModel(SimpleConcentrationModel):
    """
    Simple model for the background (long-range) exposure, without
    all the flexibility of caimira.models.ExposureModel.
    For independent, end-to-end testing purposes.
    This assumes no mask wearing, identical inhalation and exhalation
    breathing rate, indentical presence for the infected and the exposed
    people, the same ventilation rate at all times, and all short-range
    interaction intervals are within presence intervals of the infected.
    """

    #: Infectious dose (ID50)
    ID50: _VectorisedFloat = 50.

    #: Transmissibility factor w.r.t. original strain
    # (<1 means more transmissible)
    transmissibility: _VectorisedFloat = 1.

    #: List of short-range interaction models
    sr_models: typing.Tuple[SimpleShortRangeModel, ...] = ()

    def fdep(self, diameter: float, evaporation: float) -> float:
        """
        Fraction deposited
        """
        d = diameter * evaporation
        IFrac = 1 - 0.5 * (1 - (1 / (1 + (0.00076*(d**2.8)))))
        fdep = IFrac * (0.0587
                    + (0.911/(1 + np.exp(4.77 + 1.485 * np.log(d))))
                    + (0.943/(1 + np.exp(0.508 - 2.58 * np.log(d)))))
        return fdep

    @method_cache
    def F(self, removal_rate: _VectorisedFloat, deltat: float,
          evaporation: float) -> _VectorisedFloat:
        """
        A general function to compute the main integral over diameters
        """
        def integrand(diameter):
            # Function to return the integrand
            a = self.deposition_removal_coefficient()
            a_dsquare = a*diameter**2
            return -(self.vR(diameter)*self.fdep(diameter,evaporation)/(
                    (a_dsquare + removal_rate)**2)
                    * np.exp(-a_dsquare*deltat))

        return (quad(integrand,self.diameter_min,self.diameter_max,
                epsabs=0.,limit=500)[0]
                * self.viral_load * self.breathing_rate)

    @method_cache
    def f_with_fdep(self, removal_rate: _VectorisedFloat, deltat: float,
                    evaporation: float) -> _VectorisedFloat:
        """
        Same as f but with fdep included in the integral.
        """
        def integrand(diameter):
            # Function to return the integrand
            a = self.deposition_removal_coefficient()
            a_dsquare = a*diameter**2
            return (self.vR(diameter)*self.fdep(diameter,evaporation)
                    /(a_dsquare + removal_rate) * np.exp(-a_dsquare*deltat))

        return (quad(integrand,self.diameter_min,self.diameter_max,
                epsabs=0.,limit=500)[0]
                * self.viral_load * self.breathing_rate)

    def total_concentration(self, t: float):
        """
        Total concentration at time t
        """
        res = self.concentration(t)
        for sr_mod in self.sr_models:
            res += sr_mod.concentration(self,t)
        return res

    @method_cache
    def integrated_longrange_concentration(self,t1: float,t2: float,
                            evaporation: float) -> _VectorisedFloat:
        """
        Background (long-range) concentration integrated from t1 to t2
        assuming both t1 and t2 are within a single presence interval.
        This includes the deposition fraction (fdep).
        """
        trans_times = sorted(self.infected_presence.transition_times())
        if t2==trans_times[0]:
            return 0.

        lambda_rate = self.removal_rate()
        # transition_times[i] < t2 <=  transition_times[i]+1
        i: int = np.searchsorted(trans_times,t2) - 1 # type: ignore
        ti = trans_times[i]
        if np.searchsorted(trans_times,t1,side='right')-1 != i:
            raise ValueError("t1={}, t2={}, i1={}, i2={} "
                             "(i1 and i2 should be equal)".format(
                t1,t2,i,np.searchsorted(trans_times,t1,side='right')-1))

        Pim1 = (False if i==0 else
            self.infected_presence.triggered((trans_times[i-1]+ti)/2.))
        Pi = self.infected_presence.triggered((ti+trans_times[i+1])/2.)

        def primitive(time):
            result = (0 if not Pim1 else self.F(lambda_rate,time-ti,evaporation))
            result -= (0 if not Pi else self.F(lambda_rate,time-ti,evaporation))

            for k,tk in enumerate(trans_times[:i]):
                Pkm1 = (False if k==0 else
                    self.infected_presence.triggered((trans_times[k-1]+tk)/2.))
                Pk = self.infected_presence.triggered((tk+trans_times[k+1])/2.)
                s = np.sum([lambda_rate*(trans_times[l]-trans_times[l-1])
                            for l in range(k+1,i+1)])
                result += ( (0 if not Pkm1 else self.F(lambda_rate,time-tk,evaporation))
                           -(0 if not Pk else self.F(lambda_rate,time-tk,evaporation))
                           ) * np.exp(-s)
            return result

        return ( ( (0 if not self.infected_presence.triggered((t1+t2)/2.)
                    else self.f_with_fdep(lambda_rate,0,evaporation)*(t2-t1))
                  + (primitive(t2) * np.exp(-lambda_rate*(t2-ti)) -
                     primitive(t1) * np.exp(-lambda_rate*(t1-ti)) ) )
                * self.num_infected * self.fraction_of_infectious_virus()
                / self.room_volume)

    @method_cache
    def integrated_shortrange_concentration(self) -> _VectorisedFloat:
        """
        Short-range concentration integrated over interaction times and
        diameters. This includes the deposition fraction (fdep).
        """
        result = 0.
        # evaporation set to 1 (particles do not have time to evaporate)
        evaporation = 1.
        for sr_model in self.sr_models:
            t1,t2 = sr_model.interaction_interval.boundaries()[0]
            # function to return the integrand
            integrand = lambda d: (self.aerosol_volume(d)
                    * self.Np(d,sr_model.BLO_factors)*self.fdep(d,evaporation))
            res = (quad(integrand,
                        sr_model.diameter_min,sr_model.diameter_max,
                        epsabs=0.,limit=500)[0]
                   * self.viral_load * self.fraction_of_infectious_virus() * 1e-6 * (t2-t1) )
            result += sr_model.breathing_rate * (
                        res-self.integrated_longrange_concentration(t1,t2,evaporation)
                        )/sr_model.dilution_factor()

        return result

    def dose(self) -> _VectorisedFloat:
        """
        Total deposited dose (integrated over time and over particle
        diameters), including short and long-range.
        """
        result = 0.
        for t1,t2 in self.infected_presence.boundaries():
            result += (self.integrated_longrange_concentration(t1,t2,self.evaporation)
                       * self.breathing_rate)

        result += self.integrated_shortrange_concentration()

        return result

    def probability_infection(self):
        """
        Total probability of infection
        """
        return (1. - np.exp(-self.dose() * ln2 * (1-self.HI)
                           /(self.ID50 * self.transmissibility) )) * 100.


presence = models.SpecificInterval(present_times=((8.5, 12), (13, 17.5)))
interaction_intervals = (models.SpecificInterval(present_times=((10.5, 11.0),)),
                         models.SpecificInterval(present_times=((14.5, 15.0),))
                         )


@pytest.fixture
def c_model(data_registry) -> mc.ConcentrationModel:
    return mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=50, inside_temp=models.PiecewiseConstant((0., 24.), (293,)), humidity=0.3),
        ventilation=models.AirChange(active=models.PeriodicInterval(period=120, duration=120), air_exch=1.),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=presence,
            virus=models.Virus.types['SARS_CoV_2_DELTA'],
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Seated'],
            expiration=expiration_distributions(data_registry)['Breathing'],
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )


@pytest.fixture
def c_model_distr(data_registry) -> mc.ConcentrationModel:
    return mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=50, humidity=0.3),
        ventilation=models.AirChange(active=models.PeriodicInterval(
                            period=120, duration=120), air_exch=1.),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=presence,
            virus=virus_distributions(data_registry)['SARS_CoV_2_DELTA'],
            mask=models.Mask.types['No mask'],
            activity=activity_distributions(data_registry)['Seated'],
            expiration=expiration_distributions(data_registry)['Breathing'],
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )


@pytest.fixture
def sr_models(data_registry) -> typing.Tuple[mc.ShortRangeModel, ...]:
    return (
        mc.ShortRangeModel(
            data_registry = data_registry,
            expiration = short_range_expiration_distributions(data_registry)['Speaking'],
            activity = models.Activity.types['Seated'],
            presence = interaction_intervals[0],
            distance = 0.854,
        ),
        mc.ShortRangeModel(
            data_registry = data_registry,
            expiration = short_range_expiration_distributions(data_registry)['Breathing'],
            activity = models.Activity.types['Heavy exercise'],
            presence = interaction_intervals[1],
            distance = 0.854,
        ),
    )


@pytest.fixture
def simple_c_model(data_registry) -> SimpleConcentrationModel:
    return SimpleConcentrationModel(
        infected_presence = presence,
        viral_load        = models.Virus.types['SARS_CoV_2_DELTA'].viral_load_in_sputum,
        breathing_rate    = models.Activity.types['Seated'].exhalation_rate,
        room_volume       = 50.,
        lambda_ventilation= 1.,
        BLO_factors       = expiration_BLO_factors(data_registry)['Breathing'],
        viable_to_RNA     = models.Virus.types['SARS_CoV_2_DELTA'].viable_to_RNA_ratio,
        HI                = 0.,
    )


@pytest.fixture
def simple_sr_models(data_registry) -> typing.Tuple[SimpleShortRangeModel, ...]:
    return (
        SimpleShortRangeModel(
            interaction_interval = interaction_intervals[0],
            distance = 0.854,
            breathing_rate = models.Activity.types['Seated'].exhalation_rate,
            BLO_factors = expiration_BLO_factors(data_registry)['Speaking'],
        ),
        SimpleShortRangeModel(
            interaction_interval = interaction_intervals[1],
            distance = 0.854,
            breathing_rate = models.Activity.types['Heavy exercise'].exhalation_rate,
            BLO_factors = expiration_BLO_factors(data_registry)['Breathing'],
        ),
    )


@pytest.fixture
def expo_sr_model(data_registry, c_model,sr_models) -> mc.ExposureModel:
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=c_model,
        short_range=sr_models,
        exposed=mc.Population(
            number=1,
            presence=presence,
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Seated'],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def simple_expo_sr_model(data_registry, simple_sr_models) -> SimpleExposureModel:
    return SimpleExposureModel(
        infected_presence = presence,
        viral_load        = models.Virus.types['SARS_CoV_2_DELTA'].viral_load_in_sputum,
        breathing_rate    = models.Activity.types['Seated'].exhalation_rate,
        room_volume       = 50.,
        lambda_ventilation= 1.,
        BLO_factors       = expiration_BLO_factors(data_registry)['Breathing'],
        viable_to_RNA     = models.Virus.types['SARS_CoV_2_DELTA'].viable_to_RNA_ratio,
        HI                = 0.,
        ID50              = models.Virus.types['SARS_CoV_2_DELTA'].infectious_dose,
        transmissibility  = models.Virus.types['SARS_CoV_2_DELTA'].transmissibility_factor,
        sr_models         = simple_sr_models,
    )


@pytest.fixture
def expo_sr_model_distr(data_registry, c_model_distr) -> mc.ExposureModel:
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=c_model_distr,
        short_range=(
            mc.ShortRangeModel(
                data_registry = data_registry,
                expiration = short_range_expiration_distributions(data_registry)['Breathing'],
                activity = activity_distributions(data_registry)['Seated'],
                presence = interaction_intervals[0],
                distance = short_range_distances(data_registry),
            ),
            mc.ShortRangeModel(
                data_registry = data_registry,
                expiration = short_range_expiration_distributions(data_registry)['Speaking'],
                activity = activity_distributions(data_registry)['Seated'],
                presence = interaction_intervals[1],
                distance = short_range_distances(data_registry),
            ),
        ),
        exposed=mc.Population(
            number=1,
            presence=presence,
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Seated'],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@pytest.fixture
def simple_expo_sr_model_distr(data_registry) -> SimpleExposureModel:
    return SimpleExposureModel(
        infected_presence = presence,
        viral_load        = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).viral_load_in_sputum,
        breathing_rate    = activity_distributions(data_registry)['Seated'].build_model(
                                            SAMPLE_SIZE).exhalation_rate,
        room_volume       = 50.,
        lambda_ventilation= 1.,
        BLO_factors       = expiration_BLO_factors(data_registry)['Breathing'],
        viable_to_RNA     = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).viable_to_RNA_ratio,
        HI                = 0.,
        ID50              = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).infectious_dose,
        transmissibility  = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].transmissibility_factor,
        sr_models         = (
            SimpleShortRangeModel(
                interaction_interval = interaction_intervals[0],
                distance = short_range_distances(data_registry).generate_samples(SAMPLE_SIZE),
                breathing_rate = activity_distributions(data_registry)['Seated'].build_model(
                                            SAMPLE_SIZE).exhalation_rate,
                BLO_factors = expiration_BLO_factors(data_registry)['Breathing'],
            ),
            SimpleShortRangeModel(
                interaction_interval = interaction_intervals[1],
                distance = short_range_distances(data_registry).generate_samples(SAMPLE_SIZE),
                breathing_rate = activity_distributions(data_registry)['Seated'].build_model(
                                            SAMPLE_SIZE).exhalation_rate,
                BLO_factors = expiration_BLO_factors(data_registry)['Speaking'],
            )
        ),
    )


@pytest.mark.parametrize(
    "time", np.linspace(8.5,17.5,12),
)
def test_longrange_concentration(time,c_model,simple_c_model):
    npt.assert_allclose(
        c_model.build_model(SAMPLE_SIZE).concentration(time).mean(),
        simple_c_model.concentration(time), rtol=TOLERANCE
        )


@retry(tries=10)
@pytest.mark.parametrize(
    "time", [10, 10.7, 11., 12.5, 14.75, 14.9, 17]
)
def test_shortrange_concentration(time,c_model,simple_c_model,
                                  sr_models,simple_sr_models):
    result_sr_model = np.sum([np.array(
            sr_mod.build_model(SAMPLE_SIZE).short_range_concentration(c_model.build_model(SAMPLE_SIZE),time)).mean()
        for sr_mod in sr_models])
    result_simple_sr_model = np.sum([np.array(
            sr_mod.concentration(simple_c_model,time)).mean()
        for sr_mod in simple_sr_models])
    npt.assert_allclose(
        result_sr_model,result_simple_sr_model,rtol=TOLERANCE
        )


def test_longrange_exposure(data_registry, c_model):
    simple_expo_model = SimpleExposureModel(
        infected_presence = presence,
        viral_load        = models.Virus.types['SARS_CoV_2_DELTA'].viral_load_in_sputum,
        breathing_rate    = models.Activity.types['Seated'].exhalation_rate,
        room_volume       = 50.,
        lambda_ventilation= 1.,
        BLO_factors       = expiration_BLO_factors(data_registry)['Breathing'],
        viable_to_RNA     = models.Virus.types['SARS_CoV_2_DELTA'].viable_to_RNA_ratio,
        HI                = 0.,
        ID50              = models.Virus.types['SARS_CoV_2_DELTA'].infectious_dose,
        transmissibility  = models.Virus.types['SARS_CoV_2_DELTA'].transmissibility_factor,
        sr_models         = (),
    )
    expo_model = mc.ExposureModel(
            data_registry=data_registry,
            concentration_model=c_model.build_model(SAMPLE_SIZE),
            short_range=(),
            exposed=mc.Population(
                number=1,
                presence=presence,
                mask=models.Mask.types['No mask'],
                activity=models.Activity.types['Seated'],
                host_immunity=0.,
            ),
            geographical_data=models.Cases(),
    ).build_model(SAMPLE_SIZE)
    npt.assert_allclose(
        expo_model.deposited_exposure().mean(),
        simple_expo_model.dose().mean(), rtol=TOLERANCE
        )
    npt.assert_allclose(
        expo_model.infection_probability().mean(),
        simple_expo_model.probability_infection().mean(), rtol=TOLERANCE
        )


@pytest.mark.parametrize(
    "time", [11., 12.5, 17.]
)
def test_longrange_concentration_with_distributions(c_model_distr, time, data_registry):
    simple_expo_model = SimpleConcentrationModel(
        infected_presence = presence,
        viral_load        = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).viral_load_in_sputum,
        breathing_rate    = activity_distributions(data_registry)['Seated'].build_model(
                                            SAMPLE_SIZE).exhalation_rate,
        room_volume       = 50.,
        lambda_ventilation= 1.,
        BLO_factors       = expiration_BLO_factors(data_registry)['Breathing'],
        viable_to_RNA     = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).viable_to_RNA_ratio,
        HI                = 0.,
    )
    npt.assert_allclose(
        c_model_distr.build_model(SAMPLE_SIZE).concentration(time).mean(),
        simple_expo_model.concentration(time).mean(), rtol=TOLERANCE
        )


def test_longrange_exposure_with_distributions(data_registry, c_model_distr):
    simple_expo_model = SimpleExposureModel(
        infected_presence = presence,
        viral_load        = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).viral_load_in_sputum,
        breathing_rate    = activity_distributions(data_registry)['Seated'].build_model(
                                            SAMPLE_SIZE).exhalation_rate,
        room_volume       = 50.,
        lambda_ventilation= 1.,
        BLO_factors       = expiration_BLO_factors(data_registry)['Breathing'],
        viable_to_RNA     = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).viable_to_RNA_ratio,
        HI                = 0.,
        ID50              = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].build_model(SAMPLE_SIZE).infectious_dose,
        transmissibility  = virus_distributions(data_registry)['SARS_CoV_2_DELTA'
                        ].transmissibility_factor,
        sr_models         = (),
    )
    expo_model = mc.ExposureModel(
            data_registry=data_registry,
            concentration_model=c_model_distr.build_model(SAMPLE_SIZE),
            short_range=(),
            exposed=mc.Population(
                number=1,
                presence=presence,
                mask=models.Mask.types['No mask'],
                activity=activity_distributions(data_registry)['Seated'],
                host_immunity=0.,
            ),
            geographical_data=models.Cases(),
    ).build_model(SAMPLE_SIZE)
    npt.assert_allclose(
        expo_model.deposited_exposure().mean(),
        simple_expo_model.dose().mean(), rtol=TOLERANCE
        )
    npt.assert_allclose(
        expo_model.infection_probability().mean(),
        simple_expo_model.probability_infection().mean(), rtol=TOLERANCE
        )


# Tests on the concentration with short-range should be skipped until
# one finds a way to avoid the large variability of the concentration
# with short-range 'Speaking' or 'Shouting' interactions
@pytest.mark.skip
@pytest.mark.parametrize(
    "time", [10.75, 14.75, 16.]
)
def test_concentration_with_shortrange(expo_sr_model,simple_expo_sr_model,time):
    npt.assert_allclose(
        expo_sr_model.build_model(SAMPLE_SIZE).concentration(time).mean(),
        simple_expo_sr_model.total_concentration(time).mean(), rtol=TOLERANCE
        )


@retry(tries=10)
def test_exposure_with_shortrange(expo_sr_model,simple_expo_sr_model):
    npt.assert_allclose(
        expo_sr_model.build_model(SAMPLE_SIZE).deposited_exposure().mean(),
        simple_expo_sr_model.dose().mean(), rtol=TOLERANCE
        )
    npt.assert_allclose(
        expo_sr_model.build_model(SAMPLE_SIZE).infection_probability().mean(),
        simple_expo_sr_model.probability_infection().mean(), rtol=TOLERANCE
        )


@pytest.mark.skip
@pytest.mark.parametrize(
    "time", [10.75, 14.75, 16.]
)
def test_concentration_with_shortrange_and_distributions(
                    expo_sr_model_distr,simple_expo_sr_model_distr,time):
    npt.assert_allclose(
        expo_sr_model_distr.build_model(SAMPLE_SIZE).concentration(time).mean(),
        simple_expo_sr_model_distr.total_concentration(time).mean(),
        rtol=TOLERANCE
        )


@retry(tries=10)
def test_exposure_with_shortrange_and_distributions(expo_sr_model_distr,
                                            simple_expo_sr_model_distr):
    npt.assert_allclose(
        expo_sr_model_distr.build_model(SAMPLE_SIZE).deposited_exposure().mean(),
        simple_expo_sr_model_distr.dose().mean(), rtol=0.05
        )
    npt.assert_allclose(
        expo_sr_model_distr.build_model(SAMPLE_SIZE).infection_probability().mean(),
        simple_expo_sr_model_distr.probability_infection().mean(),
        rtol=0.03
        )


def exposure_model_from_parameter(data_registry, short_range_models, f_inf=0.5, viral_load=1e9, BR=1.25):
    virus: models.SARSCoV2 = models.SARSCoV2(
        viral_load_in_sputum=viral_load,
        infectious_dose=50,
        viable_to_RNA_ratio=f_inf,
        transmissibility_factor=0.51,
    )
    c_model = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=50, humidity=0.3),
        ventilation=models.AirChange(active=models.PeriodicInterval(period=120, duration=120),
                                     air_exch=10_000_000),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=models.SpecificInterval(present_times=((8.5, 12.5), (13.5, 17.5))),
            virus=virus,
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Seated'],
            expiration=expiration_distributions(data_registry)['Breathing'],
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=c_model,
        short_range=short_range_models,
        exposed=mc.Population(
            number=1,
            presence=models.SpecificInterval(present_times=((8.5, 12.5), (13.5, 17.5))),
            mask=models.Mask.types['No mask'],
            activity=models.Activity(inhalation_rate=BR, exhalation_rate=1.25),
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    ).build_model(SAMPLE_SIZE)


@retry(tries=10)
def test_exposure_scale_with_f_inf(data_registry, sr_models):
    """
    Exposure scaling test for the fraction of infectious virus.
    """
    e_model_1: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=sr_models, f_inf=0.5)
    e_model_2: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=sr_models, f_inf=1)
    np.testing.assert_allclose(
        2*e_model_1.deposited_exposure().mean(),
        e_model_2.deposited_exposure().mean(), rtol=0.02
    )


@retry(tries=10)
def test_exposure_scale_with_viral_load(data_registry, sr_models):
    """
    Exposure scaling test for the viral load.
    """
    e_model_1: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=sr_models, viral_load=1e9)
    e_model_2: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=sr_models, viral_load=2e9)
    np.testing.assert_allclose(
        2*e_model_1.deposited_exposure().mean(),
        e_model_2.deposited_exposure().mean(), rtol=0.02
    )


@retry(tries=10)
def test_lr_exposure_scale_with_breathing_rate(data_registry):
    """
    Exposure scaling test for the breathing rate when there are only long-range 
    interactions defined. Only the inhalation rate of the infected takes place 
    at the deposited exposure level.
    """
    e_model_1: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=(), BR=1.25)
    e_model_2: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=(), BR=2.5)
    np.testing.assert_allclose(
        2*e_model_1.deposited_exposure().mean(),
        e_model_2.deposited_exposure().mean(), rtol=0.02
    )


@retry(tries=10)
def test_exposure_scale_with_breathing_rate(data_registry, sr_models):
    """
    Exposure scaling test for the breathing rate when long- and short-range
    interactions are defined. We need to apply the multiplication factor
    to the inhalation rate of the infected (long-range), but also for
    each short-range interaction.
    """
    e_model_1: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=sr_models, BR=1.25)
    
    seated_act = models.Activity.types['Seated']
    heavy_exercise_act = models.Activity.types['Heavy exercise']
    sr_models_activity = (
        mc.ShortRangeModel(
            data_registry = data_registry,
            expiration = short_range_expiration_distributions(data_registry)['Speaking'],
            activity = models.Activity(inhalation_rate=seated_act.inhalation_rate * 2, 
                                       exhalation_rate=seated_act.exhalation_rate),
            presence = interaction_intervals[0],
            distance = 0.854,
        ),
        mc.ShortRangeModel(
            data_registry = data_registry,
            expiration = short_range_expiration_distributions(data_registry)['Breathing'],
            activity = models.Activity(inhalation_rate=heavy_exercise_act.inhalation_rate * 2, 
                                       exhalation_rate=heavy_exercise_act.inhalation_rate),
            presence = interaction_intervals[1],
            distance = 0.854,
        ),
    )
    e_model_2: models.ExposureModel = exposure_model_from_parameter(data_registry=data_registry, short_range_models=sr_models_activity, BR=2.5)
    np.testing.assert_allclose(
        2*e_model_1.deposited_exposure().mean(),
        e_model_2.deposited_exposure().mean(), rtol=0.02
    )



import numpy as np
import numpy.testing as npt
import pytest
from retry import retry

from caimira.calculator.models.monte_carlo import sampleable


@retry(tries=10)
@pytest.mark.parametrize(
    "mean, std",[
        [1., 0.5],
    ]
)
def test_normal(mean, std):
    # Test that the sample has approximately the right mean,
    # std deviation and distribution function.
    sample_size = 2000000
    samples = sampleable.Normal(mean, std).generate_samples(sample_size)
    histogram, bins = np.histogram(samples,bins=100, density=True)
    selected_bins,selected_histogram = zip(*[(b,h) for b,h in zip(
                (bins[1:]+bins[:-1])/2,histogram) if b>=0.25 and b<=1.75])
    exact_dist = 1/(np.sqrt(2*np.pi)*std) * np.exp(
                    -((np.array(selected_bins)-mean)/std)**2/2)

    assert len(samples) == sample_size
    npt.assert_allclose([samples.mean(), samples.std()], [mean, std], rtol=0.01)
    npt.assert_allclose(selected_histogram, exact_dist, rtol=0.02)


@pytest.mark.parametrize(
    "mean_gaussian, std_gaussian",[
        [-0.6872121723362303, 0.10498338229297108],
    ]
)
def test_lognormal(mean_gaussian, std_gaussian):
    # Test that the sample has approximately the right mean,
    # std deviation and distribution function.
    sample_size = 2000000
    samples = sampleable.LogNormal(mean_gaussian, std_gaussian
                                ).generate_samples(sample_size)
    histogram, bins = np.histogram(samples,bins=50, density=True)
    selected_bins,selected_histogram = zip(*[(b,h) for b,h in zip(
                (bins[1:]+bins[:-1])/2,histogram) if b>=0.4 and b<=0.6])
    selected_bins = np.array(selected_bins)
    exact_dist = ( 1/(selected_bins*np.sqrt(2*np.pi)*std_gaussian) *
            np.exp(-((np.log(selected_bins)-mean_gaussian)/std_gaussian)**2/2) )
    exact_mean = np.exp(mean_gaussian + std_gaussian**2/2)
    exact_std = np.sqrt( (np.exp(std_gaussian**2)-1) *
                         np.exp(2*mean_gaussian + std_gaussian**2) )

    assert len(samples) == sample_size
    npt.assert_allclose([samples.mean(), samples.std()],
                        [exact_mean, exact_std], rtol=0.01)
    npt.assert_allclose(selected_histogram, exact_dist, rtol=0.03)


@pytest.mark.parametrize(
    "use_kernel",
    [False, True],
)
def test_custom(use_kernel):
    # Test that the sample has approximately the right distribution
    # function, with both Custom and CustomKernel method. The latter
    # is less accurate for smooth functions.
    # the distribution function is an inverted parabola, with maximum 0.15,
    # which is 0 at the bounds (0,10) (normalized)
    norm = 500/3.
    function = lambda x: (-(5 - x)**2 + 25)/norm
    max_function = 0.15
    sample_size = 2000000

    if use_kernel:
        variable = np.linspace(0.1,9.9,100) 
        frequencies = function(variable)
        samples = sampleable.CustomKernel(variable, frequencies,
                                    kernel_bandwidth=0.1
                                    ).generate_samples(sample_size)
    else:
        samples = sampleable.Custom((0, 10), function, max_function
                                    ).generate_samples(sample_size)

    histogram, bins = np.histogram(samples, bins=100, density=True)
    selected_bins,selected_histogram = zip(*[(b,h) for b,h in zip(
                (bins[1:]+bins[:-1])/2,histogram) if b>=1 and b<=9])
    correct_dist = function(np.array(selected_bins))
    assert len(samples) == sample_size
    npt.assert_allclose(selected_histogram, correct_dist, rtol=0.05)


def test_logcustomkernel():
    # Test that the sample has approximately the right distribution
    # function, for the LogCustomKernel.
    # the distribution function is an inverted parabola vs. the log of
    # the variable (normalized)
    norm = 500/3.
    function = lambda x: (-(5 - x)**2 + 25)/norm
    sample_size = 2000000

    log_variable = np.linspace(0.1,9.9,100)
    frequencies = function(log_variable)
    samples = sampleable.LogCustomKernel(log_variable, frequencies,
                                kernel_bandwidth=0.1
                                ).generate_samples(sample_size)

    histogram, bins = np.histogram(np.log10(samples), bins=100, density=True)
    selected_bins,selected_histogram = zip(*[(b,h) for b,h in zip(
                (bins[1:]+bins[:-1])/2,histogram) if b>=1 and b<=9])
    correct_dist = function(np.array(selected_bins))
    assert len(samples) == sample_size
    npt.assert_allclose(selected_histogram, correct_dist, rtol=0.05)



import unittest
from unittest.mock import Mock, patch

from caimira.calculator.store.data_service import DataService


class DataServiceTests(unittest.TestCase):
    def setUp(self):
        # Set up any necessary test data or configurations
        self.data_service = DataService.create(host="https://dataservice.example.com")

    @patch("requests.get")
    def test_fetch_successful(self, mock_get):
        # Mock successful fetch response
        mock_get.return_value = Mock()
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {"data": "dummy_data"}
        data = self.data_service._fetch()

        # Assert that the data is returned correctly
        self.assertEqual(data, {"data": "dummy_data"})

        # Verify that the fetch method was called with the expected arguments
        mock_get.assert_called_once_with(
            "https://dataservice.example.com/data",
            headers={
                "Content-Type": "application/json",
            },
        )

    @patch("requests.get")
    def test_fetch_error(self, mock_get):
        # Mock fetch error response
        mock_get.return_value = Mock()
        mock_get.return_value.status_code = 500

        # Call the fetch method
        data = self.data_service._fetch()

        # Assert that the fetch method returns None in case of an error
        self.assertIsNone(data)



import dataclasses

import numpy as np
import pytest

import caimira.calculator.models
import caimira.calculator.models.models
import caimira.calculator.models.monte_carlo.sampleable

MODEL_CLASSES = [
    cls for cls in vars(caimira.calculator.models).values()
    if dataclasses.is_dataclass(cls)
]


def test_type_annotations():
    # Check that there are appropriate type annotations for all of the model
    # classes in caimira.models. Note that these must be statically defined in
    # caimira.monte_carlo, rather than being dynamically generated, in order to
    # allow the type system to be able to see their definition without needing
    # runtime execution.
    missing = []
    for cls in MODEL_CLASSES:
        if not hasattr(caimira.calculator.models.monte_carlo, cls.__name__):
            missing.append(cls.__name__)
            continue
        mc_cls = getattr(caimira.calculator.models.monte_carlo, cls.__name__)
        assert issubclass(mc_cls, caimira.calculator.models.monte_carlo.MCModelBase)

    if missing:
        msg = (
            'There are missing model implementations in caimira.monte_carlo. '
            'The following definitions are needed:\n  ' +
            '\n  '.join([f'{model} = build_mc_model(caimira.models.{model})' for model in missing])
        )
        pytest.fail(msg)


@pytest.fixture
def baseline_mc_concentration_model(data_registry) -> caimira.calculator.models.monte_carlo.ConcentrationModel:
    mc_model = caimira.calculator.models.monte_carlo.ConcentrationModel(
        data_registry=data_registry,
        room=caimira.calculator.models.monte_carlo.Room(volume=caimira.calculator.models.monte_carlo.sampleable.Normal(75, 20),
                        inside_temp=caimira.calculator.models.models.PiecewiseConstant((0., 24.), (293,))),
        ventilation=caimira.calculator.models.monte_carlo.SlidingWindow(
            data_registry=data_registry,
            active=caimira.calculator.models.models.PeriodicInterval(period=120, duration=120),
            outside_temp=caimira.calculator.models.models.PiecewiseConstant((0., 24.), (283,)),
            window_height=1.6, opening_length=0.6,
        ),
        infected=caimira.calculator.models.models.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            virus=caimira.calculator.models.models.Virus.types['SARS_CoV_2'],
            presence=caimira.calculator.models.models.SpecificInterval(((0., 4.), (5., 8.))),
            mask=caimira.calculator.models.models.Mask.types['No mask'],
            activity=caimira.calculator.models.models.Activity.types['Light activity'],
            expiration=caimira.calculator.models.models.Expiration.types['Breathing'],
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc_model


@pytest.fixture
def baseline_mc_sr_model() -> caimira.calculator.models.monte_carlo.ShortRangeModel:
    return ()


@pytest.fixture
def baseline_mc_exposure_model(data_registry, baseline_mc_concentration_model, baseline_mc_sr_model) -> caimira.calculator.models.monte_carlo.ExposureModel:
    return caimira.calculator.models.monte_carlo.ExposureModel(
        data_registry,
        baseline_mc_concentration_model,
        baseline_mc_sr_model,
        exposed=caimira.calculator.models.models.Population(
            number=10,
            presence=baseline_mc_concentration_model.infected.presence,
            activity=baseline_mc_concentration_model.infected.activity,
            mask=baseline_mc_concentration_model.infected.mask,
            host_immunity=0.,
        ),
        geographical_data=caimira.calculator.models.models.Cases(),
    )


def test_build_concentration_model(baseline_mc_concentration_model: caimira.calculator.models.monte_carlo.ConcentrationModel):
    model = baseline_mc_concentration_model.build_model(7)
    assert isinstance(model, caimira.calculator.models.models.ConcentrationModel)
    assert isinstance(model.concentration(time=0.), float)
    conc = model.concentration(time=1.)
    assert isinstance(conc, np.ndarray)
    assert conc.shape == (7, )


def test_build_exposure_model(baseline_mc_exposure_model: caimira.calculator.models.monte_carlo.ExposureModel):
    model = baseline_mc_exposure_model.build_model(7)
    assert isinstance(model, caimira.calculator.models.models.ExposureModel)
    prob = model.deposited_exposure()
    assert isinstance(prob, np.ndarray)
    assert prob.shape == (7, )



from caimira.calculator.models.dataclass_utils import nested_replace


def test_exposure_r0(baseline_exposure_model):
    baseline_n3 = nested_replace(
        baseline_exposure_model, {'concentration_model.infected.number': 3}
    )
    # The number of new cases should be greater if there are more infecteds, but
    # the reproduction number should be the same (it is a measure of one infected case).
    assert baseline_n3.expected_new_cases() > baseline_exposure_model.expected_new_cases()
    assert baseline_n3.reproduction_number() == baseline_exposure_model.reproduction_number()



import numpy as np
import pytest
from retry import retry

import caimira.calculator.models.monte_carlo as mc
from caimira.calculator.models import models
from caimira.calculator.models.dataclass_utils import nested_replace
from caimira.calculator.report import virus_report_data
from caimira.calculator.models.monte_carlo.data import activity_distributions, virus_distributions, expiration_distributions


@pytest.fixture
def baseline_exposure_model(data_registry):
    concentration_mc = mc.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=50, inside_temp=models.PiecewiseConstant((0., 24.), (298,)), humidity=0.5),
        ventilation=models.MultipleVentilation(
            ventilations=(
                models.AirChange(active=models.PeriodicInterval(period=120, duration=120), air_exch=0.25),
            )
        ),
        infected=mc.InfectedPopulation(
            data_registry=data_registry,
            number=1,
            presence=mc.SpecificInterval(present_times=((0, 3.5), (4.5, 9))),
            virus=virus_distributions(data_registry)['SARS_CoV_2_DELTA'],
            mask=models.Mask.types['No mask'],
            activity=activity_distributions(data_registry)['Seated'],
            expiration=expiration_distributions(data_registry)['Breathing'],
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return mc.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_mc,
        short_range=(),
        exposed=mc.Population(
            number=3,
            presence=mc.SpecificInterval(present_times=((0, 3.5), (4.5, 9))),
            activity=activity_distributions(data_registry)['Seated'],
            mask=models.Mask.types['No mask'],
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


@retry(tries=3)
def test_conditional_prob_inf_given_vl_dist(data_registry, baseline_exposure_model):

    viral_loads = np.array([3., 5., 7., 9.,])
    mc_model: models.ExposureModel = baseline_exposure_model.build_model(2_000_000)

    expected_pi_means = []
    expected_lower_percentiles = []
    expected_upper_percentiles = []

    for vl in viral_loads:
        model_vl: models.ExposureModel = nested_replace(
            mc_model, {
                'concentration_model.infected.virus.viral_load_in_sputum' : 10**vl,
            }
        )
        pi = model_vl.infection_probability()/100

        expected_pi_means.append(np.mean(pi))
        expected_lower_percentiles.append(np.quantile(pi, 0.05))
        expected_upper_percentiles.append(np.quantile(pi, 0.95))

    infection_probability = mc_model.infection_probability() / 100
    specific_vl = np.log10(mc_model.concentration_model.infected.virus.viral_load_in_sputum)
    step = 8/100
    actual_pi_means, actual_lower_percentiles, actual_upper_percentiles = (
        virus_report_data.conditional_prob_inf_given_vl_dist(infection_probability, viral_loads, specific_vl, step)
    )

    assert np.allclose(actual_pi_means, expected_pi_means, atol=0.002)
    assert np.allclose(actual_lower_percentiles, expected_lower_percentiles, atol=0.002)
    assert np.allclose(actual_upper_percentiles, expected_upper_percentiles, atol=0.002)



import json
import traceback
import tornado.web
import sys

from caimira.api.controller.virus_report_controller import submit_virus_form
from caimira.api.controller.co2_report_controller import submit_CO2_form


class BaseReportHandler(tornado.web.RequestHandler):
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")

    def write_error(self, status_code, **kwargs):
        self.set_status(status_code)
        self.write({"message": kwargs.get('exc_info')[1].__str__()})


class VirusReportHandler(BaseReportHandler):
    def post(self):
        try:
            form_data = json.loads(self.request.body)
            arguments = self.request.arguments
            # Report generation parallelism argument
            try:
                report_generation_parallelism = int(arguments['report_generation_parallelism'][0])
            except (ValueError, IndexError, KeyError):
                report_generation_parallelism = None

            report_data = submit_virus_form(form_data, report_generation_parallelism)

            response_data = {
                "status": "success",
                "message": "Results generated successfully",
                "report_data": report_data,
            }

            self.write(response_data)
        except Exception as e:
            traceback.print_exc()
            self.write_error(status_code=400, exc_info=sys.exc_info())


class CO2ReportHandler(BaseReportHandler):
    def post(self):
        try:
            form_data = json.loads(self.request.body)
            report_data = submit_CO2_form(form_data)

            response_data = {
                "status": "success",
                "message": "Results generated successfully",
                "report_data": report_data,
            }

            self.write(response_data)
        except Exception as e:
            traceback.print_exc()
            self.write_error(status_code=400, exc_info=sys.exc_info())



import numpy as np
import numpy.testing as npt
import pytest

from caimira.calculator.models.monte_carlo.data import activity_distributions, virus_distributions


# Mean & std deviations from https://doi.org/10.1101/2021.10.14.21264988 (Table 3)
# NOTE: a mistake was corrected for the std deviation of the
# "Moderate exercise" case (0.37 in the report, but should be 0.34)
@pytest.mark.parametrize(
    "distribution, mean, std",[
        ['Seated', 0.51, 0.053],

        ['Standing', 0.57, 0.053],

        ['Light activity', 1.24, 0.12],

        ['Moderate activity', 1.77, 0.34],

        ['Heavy exercise', 3.28, 0.72,],
    ]
)
def test_activity_distributions(data_registry, distribution, mean, std):
    activity = activity_distributions(data_registry)[distribution].build_model(size=1000000)
    npt.assert_allclose(activity.inhalation_rate.mean(), mean, atol=0.01)
    npt.assert_allclose(activity.inhalation_rate.std(), std, atol=0.01)


# Mean & std deviations from https://doi.org/10.1101/2021.10.14.21264988 (Table 3)
# - with a refined precision on the values
@pytest.mark.parametrize(
    "distribution, mean, std",[
        ['SARS_CoV_2', 6.22, 1.80],

        ['SARS_CoV_2_ALPHA', 6.22, 1.80],

        ['SARS_CoV_2_GAMMA', 6.22, 1.80],
    ]
)
def test_viral_load_logdistribution(data_registry, distribution, mean, std):
    virus = virus_distributions(data_registry)[distribution].build_model(size=1000000)
    npt.assert_allclose(np.log10(virus.viral_load_in_sputum).mean(), mean, atol=0.01)
    npt.assert_allclose(np.log10(virus.viral_load_in_sputum).std(), std, atol=0.01)



import dataclasses

from caimira.calculator.models.dataclass_utils import nested_replace, walk_dataclass


@dataclasses.dataclass(frozen=True)
class Four:
    four: float


@dataclasses.dataclass(frozen=True)
class Two:
    three: int
    four: Four


@dataclasses.dataclass(frozen=True)
class One:
    one: int
    two: Two



def test_nested_replace():
    inst = One(1, two=Two(3, Four(4)))
    new_inst = nested_replace(inst, {'two.four': Four(5)})
    assert new_inst == One(1, two=Two(3, Four(5)))


def test_walk():
    inst = One(1, two=Two(3, Four(4)))
    expected = [
        ('inst.one', inst.one),
        ('inst.two', inst.two),
        ('inst.two.three', inst.two.three),
        ('inst.two.four', inst.two.four),
        ('inst.two.four.four', inst.two.four.four),
    ]
    assert list(walk_dataclass(inst, name='inst')) == expected



import re

import numpy as np
import numpy.testing as npt
import pytest
from retry import retry

from caimira.calculator.models import models
from caimira.calculator.models.monte_carlo.data import expiration_distribution


def test_multiple_wrong_weight_size():
    weights = (1., 2., 3.)
    e_base = models.Expiration(2.5)
    with pytest.raises(
            ValueError,
            match=re.escape("expirations and weigths must contain the"
                            "same number of elements")
    ):
        e = models.MultipleExpiration([e_base, e_base], weights)


def test_multiple_wrong_diameters():
    weights = (1., 2., 3.)
    e1 = models.Expiration(np.array([1., 1.]))
    e2 = models.Expiration(1.)
    e3 = models.Expiration(2.)
    with pytest.raises(
            ValueError,
            match=re.escape("diameters must all be scalars")
    ):
        e = models.MultipleExpiration([e1, e2, e3], weights)


def test_multiple():
    weights = (1., 1.)
    mask = models.Mask.types['Type I']
    e1 = models.Expiration.types['Breathing']
    e2 = models.Expiration.types['Singing']
    aerosol_expected = (e1.aerosols(mask) + e2.aerosols(mask))/2.
    e = models.MultipleExpiration([e1, e2], weights)
    npt.assert_almost_equal(aerosol_expected, e.aerosols(mask))


@retry(tries=10)
# Expected values obtained from analytical formulas
@pytest.mark.parametrize(
    "BLO_weights, expected_aerosols",
    [
        [(1.,0.,0.), 8.33551e-13],
        [(1.,1.,1.), 2.20071e-11],
        [(1.,5.,5.), 1.06701e-10],
    ],
)
def test_expiration_aerosols(data_registry, BLO_weights, expected_aerosols):
    mask = models.Mask.types['No mask']
    e = expiration_distribution(data_registry, BLO_weights)
    npt.assert_allclose(e.build_model(100000).aerosols(mask).mean(),
                        expected_aerosols, rtol=1e-2)



import numpy as np
import numpy.testing as npt
import pytest

import caimira.calculator.models.models as models
import caimira.calculator.models.data as data


def test_no_mask_superspeading_emission_rate(baseline_concentration_model):
    expected_rate = 48500.
    npt.assert_allclose(
        [baseline_concentration_model.infected.emission_rate(float(t)) for t in [0, 1, 4, 4.5, 5, 8, 9]],
        [0, expected_rate, expected_rate, 0, 0, expected_rate, 0],
        rtol=1e-12
    )


@pytest.fixture
def baseline_periodic_window(data_registry):
    return models.SlidingWindow(
        data_registry=data_registry,
        active=models.PeriodicInterval(period=120, duration=15),
        outside_temp=models.PiecewiseConstant((0., 24.), (283,)),
        window_height=1.6, opening_length=0.6,
    )


@pytest.fixture
def baseline_room():
    return models.Room(volume=75, inside_temp=models.PiecewiseConstant((0., 24.), (293,)))


@pytest.fixture
def baseline_periodic_hepa():
    return models.HEPAFilter(
        active=models.PeriodicInterval(period=120, duration=15),
        q_air_mech=514.74,
    )


def test_concentrations(baseline_concentration_model):
    # Expected concentrations were computed analytically
    ts = [0, 4, 5, 7, 10]
    concentrations = [baseline_concentration_model.concentration(float(t)) for t in ts]
    npt.assert_allclose(
        concentrations,
        [0.000000e+00, 2.046096e+01, 3.846725e-13, 2.046096e+01, 7.231966e-27],
        rtol=1e-6
    )


def test_smooth_concentrations(baseline_concentration_model):
    # We don't care about the actual concentrations in this test, but rather
    # that the curve itself is smooth.
    dx = 0.002
    dy_limit = 0.2  # Anything more than this (in relative) is a bit steep.
    ts = np.arange(0, 10, dx)
    concentrations = [baseline_concentration_model.concentration(float(t)) for t in ts]
    assert np.abs(np.diff(concentrations)).max()/np.mean(concentrations) < dy_limit


def build_model(data_registry, interval_duration):
    model = models.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=75),
        ventilation=models.HEPAFilter(
            active=models.PeriodicInterval(period=120, duration=interval_duration),
            q_air_mech=500.,
        ),
        infected=models.EmittingPopulation(
            data_registry=data_registry,
            number=1,
            virus=models.Virus.types['SARS_CoV_2'],
            presence=models.SpecificInterval(((0., 4.), (5., 8.))),
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Light activity'],
            known_individual_emission_rate=970 * 50,
            host_immunity=0.,
            # Superspreading event, where ejection factor is fixed based
            # on Miller et al. (2020) - 50 represents the infectious dose.
        ),
        evaporation_factor=0.3,
    )
    return model


def test_concentrations_startup(data_registry):
    # The concentrations should be the same until the beginning of the
    # first time that the ventilation is disabled.
    m1 = build_model(data_registry, interval_duration=120)
    m2 = build_model(data_registry, interval_duration=65)

    assert m1.concentration(1.) == m2.concentration(1.)


def test_r0(baseline_exposure_model):
    # Expected r0 was computed with a trapezoidal integration, using
    # a mesh of 100'000 pts per exposed presence interval.
    r0 = baseline_exposure_model.reproduction_number()
    npt.assert_allclose(r0, 771.380385)


def test_periodic_window(baseline_periodic_window, baseline_room):
    # Interesting transition times for a period of 120 and duration of 15.
    ts = [0, 14/60, 15/60, 16/60, 119/60, 120/60, 121/60, 239/60, 240/60]
    aes = [baseline_periodic_window.air_exchange(baseline_room, t) for t in ts]
    rate = 6.86347
    answers = [0, rate, rate, 0, 0, 0, rate, 0, 0]
    npt.assert_allclose(aes, answers, rtol=1e-5)


def test_periodic_hepa(baseline_periodic_hepa, baseline_room):
    # Interesting transition times for a period of 120 and duration of 15.
    ts = [0, 14 / 60, 15 / 60, 16 / 60, 119 / 60, 120 / 60, 121 / 60, 239 / 60, 240 / 60]
    rate = 514.74 / 75
    aes = [baseline_periodic_hepa.air_exchange(baseline_room, t) for t in ts]
    answers = [0, rate, rate, 0, 0, 0, rate, 0, 0]
    npt.assert_allclose(aes, answers, rtol=1e-5)


@pytest.mark.parametrize(
    "time, expected_value",
    [
        [0., 0.],
        [1 / 60, 514.74 / 68],
        [14 / 60, 514.74 / 68 + 5.3842316],
        [15 / 60, 514.74 / 68 + 5.3842316],
        [16 / 60, 5.3842316],
        [1., 5.3842316],
        [1.5, 3.7393925],
        [121 / 60, 514.74 / 68 + 3.7393925],
        [2.5, 3.7393925],
    ],
)
def test_multiple_ventilation_HEPA_window(data_registry, baseline_periodic_hepa, time, expected_value):
    room = models.Room(volume=68., inside_temp=models.PiecewiseConstant((0., 24.),(293.15,)))
    tempOutside = models.PiecewiseConstant((0., 1., 2.5),(273.15, 283.15))
    window = models.SlidingWindow(
        data_registry=data_registry,
        active=models.SpecificInterval([(1 / 60, 24.)]),
        outside_temp=tempOutside,
        window_height=1.,opening_length=0.6
    )
    vent = models.MultipleVentilation([window, baseline_periodic_hepa])
    npt.assert_allclose(vent.air_exchange(room,time), expected_value, rtol=1e-5)


def test_multiple_ventilation_HEPA_window_transitions(data_registry, baseline_periodic_hepa):
    tempOutside = models.PiecewiseConstant((0., 1., 2.5),(273.15, 283.15))
    room = models.Room(68, models.PiecewiseConstant((0., 24.),(293.15,)))
    window = models.SlidingWindow(
        data_registry=data_registry,
        active=models.SpecificInterval([(1 / 60, 24.)]),
        outside_temp=tempOutside,
        window_height=1.,opening_length=0.6
    )
    vent = models.MultipleVentilation([window, baseline_periodic_hepa])
    assert set(vent.transition_times(room)) == set([0.0, 1/60, 0.25, 1.0, 2.0, 2.25,
            2.5, 4.0, 4.25, 6.0, 6.25, 8.0, 8.25, 10.0, 10.25, 12.0, 12.25,
            14.0, 14.25, 16.0, 16.25, 18.0, 18.25, 20.0, 20.25, 22.0, 22.25, 24.])


@pytest.mark.parametrize(
    "volume, expected_value",
    [
        [24.5, 500 / 24.5 + 100 / 24.5 + 3.],
        [70, 500 / 70 + 100 / 70 + 3.],
    ],
)
def test_multiple_ventilation_HEPA_HVAC_AirChange(volume, expected_value):
    room = models.Room(volume=volume)
    hepa = models.HEPAFilter(
        active=models.SpecificInterval(((0,24),)),
        q_air_mech=500.,
    )
    hvac = models.HVACMechanical(
        active=models.SpecificInterval(((0,24),)),
        q_air_mech=100.,
    )
    airchange = models.AirChange(
        active=models.SpecificInterval(((0,24),)),
        air_exch=3.,
    )
    vent = models.MultipleVentilation([hepa, hvac, airchange])
    npt.assert_allclose(vent.air_exchange(room,10.),
                        expected_value,rtol=1e-5)


@pytest.mark.parametrize(
    "time, expected_value",
    [
        [8., 5.3842316],
        [16., 3.7393925],
    ],
)
def test_windowopening(data_registry, time, expected_value):
    tempOutside = models.PiecewiseConstant((0., 10., 24.),(273.15, 283.15))
    w = models.SlidingWindow(
        data_registry=data_registry,
        active=models.SpecificInterval([(0., 24.)]),
        outside_temp=tempOutside,
        window_height=1., opening_length=0.6,
    )
    npt.assert_allclose(
        w.air_exchange(models.Room(volume=68, inside_temp=models.PiecewiseConstant((0., 24.), (293.15, ))), time), expected_value, rtol=1e-5
    )


def build_hourly_dependent_model(
        data_registry,
        month,
        intervals_open=((7.5, 8.5),),
        intervals_presence_infected=((0., 4.), (5., 7.5)),
        artificial_refinement=False,
        temperatures=data.GenevaTemperatures_hourly
):
    if artificial_refinement:
        # 5-fold increase of number of times, WITHOUT interpolation
        # (hence transparent for the results)
        refine_factor = 2
        times_refined = tuple(
            float(t) for t in np.linspace(
                0., 24, refine_factor * len(temperatures[month].values) + 1
            )
        )
        temperatures_refined = tuple(np.hstack(
            [[v] * refine_factor for v in temperatures[month].values]
        ))
        outside_temp = models.PiecewiseConstant(times_refined, temperatures_refined)
    else:
        outside_temp = temperatures[month]

    model = models.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=75, inside_temp=models.PiecewiseConstant((0., 24.), (293, ))),
        ventilation=models.SlidingWindow(
            data_registry=data_registry,
            active=models.SpecificInterval(intervals_open),
            outside_temp=outside_temp,
            window_height=1.6, opening_length=0.6,
        ),
        infected=models.EmittingPopulation(
            data_registry=data_registry,
            number=1,
            virus=models.Virus.types['SARS_CoV_2'],
            presence=models.SpecificInterval(intervals_presence_infected),
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Light activity'],
            known_individual_emission_rate=970 * 50,
            host_immunity=0,
        ),
        evaporation_factor=0.3,
    )
    return model


def build_constant_temp_model(data_registry, outside_temp, intervals_open=((7.5, 8.5),)):
    model = models.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=75, inside_temp=models.PiecewiseConstant((0., 24.), (293,))),
        ventilation=models.SlidingWindow(
            data_registry=data_registry,
            active=models.SpecificInterval(intervals_open),
            outside_temp=models.PiecewiseConstant((0., 24.), (outside_temp,)),
            window_height=1.6, opening_length=0.6,
        ),
        infected=models.EmittingPopulation(
            data_registry=data_registry,
            number=1,
            virus=models.Virus.types['SARS_CoV_2'],
            presence=models.SpecificInterval(((0., 4.), (5., 7.5))),
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Light activity'],
            known_individual_emission_rate=970 * 50,
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return model


def build_hourly_dependent_model_multipleventilation(data_registry, month, intervals_open=((7.5, 8.5),)):
    vent = models.MultipleVentilation((
        models.SlidingWindow(
            data_registry=data_registry,
            active=models.SpecificInterval(intervals_open),
            outside_temp=data.GenevaTemperatures[month],
            window_height=1.6, opening_length=0.6,
        ),
        models.HEPAFilter(
            active=models.SpecificInterval(((0., 24.),)),
            q_air_mech=500.,
        ),
    ))
    model = models.ConcentrationModel(
        data_registry=data_registry,
        room=models.Room(volume=75, inside_temp=models.PiecewiseConstant((0., 24.), (293,))),
        ventilation=vent,
        infected=models.EmittingPopulation(
            data_registry=data_registry,
            number=1,
            virus=models.Virus.types['SARS_CoV_2'],
            presence=models.SpecificInterval(((0., 4.), (5., 7.5))),
            mask=models.Mask.types['No mask'],
            activity=models.Activity.types['Light activity'],
            known_individual_emission_rate=970 * 50,
            host_immunity=0.,
        ),
        evaporation_factor=0.3,
    )
    return model


@pytest.mark.parametrize(
    "month, temperatures",
    data.local_hourly_temperatures_celsius_per_hour.items(),
)
@pytest.mark.parametrize(
    "time",
    [0.5, 1.2, 2., 3.5, 5., 6.5, 7.5, 7.9, 8.],
)
def test_concentrations_hourly_dep_temp_vs_constant(data_registry, month, temperatures, time):
    # The concentrations should be the same up to 8 AM (time when the
    # temperature changes DURING the window opening).
    m1 = build_hourly_dependent_model(data_registry, month)
    m2 = build_constant_temp_model(data_registry, temperatures[7] + 273.15)
    npt.assert_allclose(m1.concentration(time), m2.concentration(time), rtol=1e-5)

@pytest.mark.parametrize(
    "month, temperatures",
    data.local_hourly_temperatures_celsius_per_hour.items(),
)
@pytest.mark.parametrize(
    "time",
    [0.5, 1.2, 2., 3.5, 5., 6.5, 7.5, 7.9, 8.],
)
def test_concentrations_hourly_dep_temp_startup(data_registry, month, temperatures, time):
    # The concentrations should be the zero up to the first presence time
    # of an infected person.
    m = build_hourly_dependent_model(
        data_registry,
        month,
        ((0., 0.5), (1., 1.5), (4., 4.5), (7.5, 8), ),
        ((8., 12.), ),
    )
    assert m.concentration(time) == 0.


def test_concentrations_hourly_dep_multipleventilation(data_registry):
    m = build_hourly_dependent_model_multipleventilation(data_registry, 'Jan')
    m.concentration(12.)


@pytest.mark.parametrize(
    "month_temp_item",
    data.local_hourly_temperatures_celsius_per_hour.items(),
)
@pytest.mark.parametrize(
    "time",
    [0.5, 1.2, 2., 3.5, 5., 6.5, 7.5, 7.9, 8., 8.5, 9., 12.],
)
def test_concentrations_hourly_dep_adding_artificial_transitions(data_registry, month_temp_item, time):
    month, temperatures = month_temp_item
    # Adding a second opening inside the first one should not change anything
    m1 = build_hourly_dependent_model(data_registry, month, intervals_open=((7.5, 8.5), ))
    m2 = build_hourly_dependent_model(data_registry, month, intervals_open=((7.5, 8.5), (8., 8.1), ))
    npt.assert_allclose(m1.concentration(time), m2.concentration(time), rtol=1e-5)


@pytest.mark.parametrize(
    "time",
    (
        [float(t) for t in np.random.random_sample(10) * 24.]  # type: ignore
        + [float(t) for t in np.arange(0, 24.5, 0.5)]
    ),
)
def test_concentrations_refine_times(data_registry, time):
    month = 'Jan'
    m1 = build_hourly_dependent_model(data_registry, month, intervals_open=((0., 24.),))
    m2 = build_hourly_dependent_model(data_registry, month, intervals_open=((0., 24.),),
                                      artificial_refinement=True)
    npt.assert_allclose(m1.concentration(time), m2.concentration(time), rtol=1e-8)


def build_exposure_model(data_registry, concentration_model, short_range_model):
    infected = concentration_model.infected
    return models.ExposureModel(
        data_registry=data_registry,
        concentration_model=concentration_model,
        short_range=short_range_model,
        exposed=models.Population(
            number=10,
            presence=infected.presence,
            activity=infected.activity,
            mask=infected.mask,
            host_immunity=0.,
        ),
        geographical_data=models.Cases(),
    )


# Expected deposited exposure were computed with a trapezoidal integration, using
# a mesh of 100'000 pts per exposed presence interval.
@pytest.mark.parametrize(
    "month, expected_deposited_exposure",
    [
        ['Jan', 359.140499],
        ['Jun', 1385.917562],
    ],
)
def test_exposure_hourly_dep(data_registry, month, expected_deposited_exposure, baseline_sr_model):
    m = build_exposure_model(
        data_registry,
        build_hourly_dependent_model(
            data_registry,
            month,
            intervals_open=((0., 24.), ),
            intervals_presence_infected=((8., 12.), (13., 17.))
        ), baseline_sr_model
    )
    deposited_exposure = m.deposited_exposure()
    npt.assert_allclose(deposited_exposure, expected_deposited_exposure)

# Expected deposited exposure were computed with a trapezoidal integration, using
# a mesh of 100'000 pts per exposed presence interval and 25 pts per hour
# for the temperature discretization.
@pytest.mark.parametrize(
    "month, expected_deposited_exposure",
    [
        ['Jan', 359.983716],
        ['Jun', 1439.267381],
    ],
)
def test_exposure_hourly_dep_refined(data_registry, month, expected_deposited_exposure, baseline_sr_model):
    m = build_exposure_model(
        data_registry,
        build_hourly_dependent_model(
            data_registry,
            month,
            intervals_open=((0., 24.),),
            intervals_presence_infected=((8., 12.), (13., 17.)),
            temperatures=data.GenevaTemperatures,
        ), baseline_sr_model
    )
    deposited_exposure = m.deposited_exposure()
    npt.assert_allclose(deposited_exposure, expected_deposited_exposure, rtol=0.02)
